#define _GNU_SOURCE
#define _WIN32_WINNT 0x0601
#define __USE_MINGW_ANSI_STDIO 1

#define VLC_SSE
#define VLC_MMX

#define HAVE_MMX2
#define HAVE_SSE

#ifdef __WIN32
#include <windows.h>
#endif

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <stdbool.h>

#include "fastmemcpy.h"

static size_t n = 1920 * 1080 * 2; // transfer size

static int64_t d(void) // timestamp
{
#ifdef __WIN32
    ULONGLONG ts;
    QueryUnbiasedInterruptTime (&ts);
    return ts / 10; /* hundreds of nanoseconds */
#else
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (int64_t)tv.tv_sec * 1000000 + tv.tv_usec;
#endif
}

static void *stupid(void *__restrict__ b, const void *__restrict__ a, size_t n)
{
    char *s1 = b;
    const char *s2 = a;
    for(; 0<n; --n)*s1++ = *s2++;
    return b;
}

static int64_t bench (void *(*cpy)(void *, const void *, size_t), size_t len, bool align)
{
    size_t log_align = 13; // 8K
    size_t alignement = (1 << alignement) - 1;

    char *to = malloc(len + alignement);
    char *from = malloc(len + alignement);
    char *a = (char*)(((uintptr_t)to + alignement) & ~alignement);
    char *b = (char*)(((uintptr_t)from + alignement) & ~alignement);

    int64_t now = d();

    size_t loops = 1000; // repeated iterations
    while (--loops)
        cpy(&b[!align], &a[!align * 2], len - !align * 2);

    int64_t ret = d() - now;

    free(to);
    free(from);

    if (ret == 0) {
        printf("Time measurement incorrect!\n");
        ret = 1;
    }

    return ret;
}


static float cmp(size_t size, bool align) // difference between vlc and glibc
{
    size_t glibc = bench(memcpy, size, align);
    size_t vlc = bench(fast_memcpy, size, align);

    return (float)vlc / glibc;
}

int main(int argc, char *argv[])
{
    if (argc == 2)
        n = atoi(argv[1]); // set block size
    if (n < 2)
        n = 2;

    printf("\tblock size %zu\n", n);

    int j;
    for (j = 0; j < 2; j++) {
        float f = cmp(n, j);
        bool faster = f > 1.0;
        if (!faster)
            f = 1.0 / f;

        printf("%saligned: libc %6.2f%% %ser than vlc.\n",
            j ? "  " : "un",
            f * 100 - 100,
            faster ? "fast" : "slow"
            );
    }

    return 0;
}
